"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const base_1 = require("@jupyter-widgets/base");
//import {Document, DocumentChangedEvent, ModelChangedEvent} from "document"
//import {Receiver, Fragment} from "protocol/receiver"
//import {keys, values} from "core/util/object"
const metadata_1 = require("./metadata");
function bk_require(name) {
    return window.Bokeh.require(name);
}
const { keys, values } = Object;
const version_range = `^${metadata_1.version}`;
class BokehModel extends base_1.DOMWidgetModel {
    defaults() {
        return Object.assign(Object.assign({}, super.defaults()), { _model_name: "BokehModel", _model_module: metadata_1.name, _model_module_version: version_range, _view_name: "BokehView", _view_module: metadata_1.name, _view_module_version: version_range, combine_events: false, render_bundle: {} });
    }
}
exports.BokehModel = BokehModel;
BokehModel.serializers = Object.assign({}, base_1.DOMWidgetModel.serializers);
class BokehView extends base_1.DOMWidgetView {
    constructor(options) {
        super(options);
        this._document = null;
        this._blocked = false;
        this._idle = true;
        this._combine = true;
        this._msgs = [];
        const { Receiver } = bk_require("protocol/receiver");
        this._receiver = new Receiver();
        this.model.on("change:render_bundle", () => this.render());
        if (window.Jupyter != null && window.Jupyter.notebook != null) {
            // Handle classic Jupyter notebook
            const events = window.require('base/js/events');
            events.on('kernel_idle.Kernel', () => this._process_msg());
        }
        else if (this.model.widget_manager.context != null) {
            // Handle JupyterLab
            this.model.widget_manager.context.sessionContext.statusChanged.connect((_, status) => {
                if (status === "idle")
                    this._process_msg();
            });
        }
        else {
            if (this.model.get("combine_events"))
                console.warn("BokehView cannot combine events because Kernel idle status cannot be determined.");
            this._combine = false;
        }
        this.listenTo(this.model, "msg:custom", (content, buffers) => this._consume_patch(content, buffers));
    }
    _process_msg() {
        if (this._msgs.length == 0) {
            this._idle = true;
            return;
        }
        this.send(this._msgs.shift());
    }
    render() {
        const bundle = JSON.parse(this.model.get("render_bundle"));
        const { docs_json, render_items, div } = bundle;
        this.el.innerHTML = div;
        const element = this.el.children[0];
        const json = values(docs_json)[0];
        const { Document } = bk_require("document");
        const { add_document_standalone } = bk_require("embed/standalone");
        this._document = Document.from_json(json);
        for (const item of render_items) {
            const roots = {};
            for (const root_id in item.roots)
                roots[root_id] = element;
            add_document_standalone(this._document, element, roots);
        }
        this._document.on_change((event) => this._change_event(event));
    }
    _combine_events(new_msg) {
        const new_msgs = [];
        for (const msg of this._msgs) {
            if (new_msg.kind != msg.kind)
                new_msgs.push(msg);
            else if (msg.kind == "ModelChanged" && new_msg.kind == "ModelChanged") {
                if (msg.id != new_msg.id || msg.attr != new_msg.attr)
                    new_msgs.push(msg);
            }
            else if (msg.kind == "MessageSent" && new_msg.kind == "MessageSent") {
                if (msg.msg_data.event_values.model.id != new_msg.msg_data.event_values.model.id ||
                    msg.msg_data.event_name != new_msg.msg_data.event_name)
                    new_msgs.push(msg);
            }
        }
        new_msgs.push(new_msg);
        return new_msgs;
    }
    _send(msg) {
        if (!this._idle && this._combine && this.model.get("combine_events"))
            // Queue event and drop previous events on same model attribute
            this._msgs = this._combine_events(msg);
        else {
            this._idle = false;
            this.send(msg);
        }
    }
    _change_event(event) {
        if (this._blocked) {
            return;
        }
        const { ModelChangedEvent, MessageSentEvent } = bk_require("document/events");
        if (event instanceof ModelChangedEvent) {
            const js_msg = {
                event: "jsevent",
                kind: "ModelChanged",
                id: event.model.id,
                attr: event.attr,
                new: event.new_,
                old: event.old,
                hint: null,
            };
            if (event.hint != null) {
                if (event.hint.patches != null) {
                    js_msg["hint"] = {
                        column_source: event.hint.column_source,
                        patches: event.hint.patches,
                    };
                }
                else if (event.hint.data != null) {
                    js_msg["hint"] = {
                        column_source: event.hint.column_source,
                        data: event.hint.data,
                        rollover: event.hint.rollover,
                    };
                }
            }
            this._send(js_msg);
        }
        else if ((event instanceof MessageSentEvent) && (event.msg_type == "bokeh_event")) {
            const msg_data = Object.assign({}, event.msg_data);
            const event_values = Object.assign({}, msg_data.event_values);
            event_values["model"] = { id: event_values.model.id };
            msg_data["event_values"] = event_values;
            const js_msg = {
                event: "jsevent",
                kind: "MessageSent",
                msg_type: event.msg_type,
                msg_data: msg_data,
            };
            this._send(js_msg);
        }
    }
    _consume_patch(content, buffers) {
        if (this._document == null)
            return;
        if (content.msg == "patch") {
            const { payload } = content;
            this._receiver.consume(payload != null ? payload : buffers[0].buffer);
            const comm_msg = this._receiver.message;
            if (comm_msg != null && keys(comm_msg.content).length > 0) {
                this._blocked = true;
                try {
                    this._document.apply_json_patch(comm_msg.content, comm_msg.buffers);
                }
                finally {
                    this._blocked = false;
                }
            }
        }
    }
}
exports.BokehView = BokehView;
